# coding: utf8
from __future__ import unicode_literals, print_function, division

import attr
from lingpy import rc

from pylexibank.lingpy_util import test_sequences, TranscriptionAnalysis
from pylexibank.util import get_variety_id, pb


@attr.s
class TranscriptionStats(TranscriptionAnalysis):
    inventory_size = attr.ib(default=0)
    invalid_words = attr.ib(default=attr.Factory(list))
    invalid_words_count = attr.ib(default=0)
    bad_words = attr.ib(default=attr.Factory(list))
    bad_words_count = attr.ib(default=0)


def analyze(dataset):
    rc(schema=dataset.metadata.lingpy_schema or 'ipa')
    ans, bad_words, invalid_words = test_sequences(
        dataset.cldf,
        get_variety_id,
        model='dolgo')

    stats = TranscriptionStats(
        bad_words=sorted(bad_words[:100], key=lambda x: x['ID']),
        bad_words_count=len(bad_words),
        invalid_words=sorted(invalid_words[:100], key=lambda x: x['ID']),
        invalid_words_count=len(invalid_words))
    for lid, analysis in ans.items():
        for attribute in ['segments', 'lingpy_errors', 'clpa_errors', 'replacements']:
            getattr(stats, attribute).update(getattr(analysis, attribute))
        stats.general_errors += analysis.general_errors
        stats.inventory_size += len(analysis.segments) / len(ans)

    error_segments = stats.lingpy_errors.union(stats.clpa_errors)
    for i, row in pb(enumerate(stats.bad_words)):
        analyzed_segements = []
        for s in row['Segments']:
            analyzed_segements.append('<s> %s </s>' % s if s in error_segments else s)
        stats.bad_words[i] = [
            row['ID'],
            row.get('Language_name', get_variety_id(row)),
            row.get('Parameter_name', row['Parameter_ID']),
            row['Form'],
            ' '.join(analyzed_segements)]

    for i, row in enumerate(stats.invalid_words):
        stats.invalid_words[i] = [
            row['ID'],
            row.get('Language_name', get_variety_id(row)),
            row.get('Parameter_name', row['Parameter_ID']),
            row['Form']]

    return dict(
        by_language={k: attr.asdict(v) for k, v in ans.items()}, stats=attr.asdict(stats))



# coding: utf8
from __future__ import unicode_literals, print_function, division

from pylexibank.util import jsondump
from pylexibank.commands.analyze import transcription


def analyze(ds, **kw):
    """
    Analyzes a dataset.

    lexibank analyze DATASET_ID
    """
    jsondump(
        transcription.analyze(ds),
        ds.dir.joinpath('transcription.json'),
        log=kw.get('log'))



# coding: utf8
from __future__ import unicode_literals, print_function, division

import attr
from csvw.metadata import Column
from clldutils.path import copy, Path
from clldutils.misc import slug
from pycldf.dataset import Wordlist


class Dataset(object):
    def __init__(self, dataset):
        self._count = 0
        self._cognate_count = 0
        self.dataset = dataset

        md = self.dataset.cldf_dir.joinpath('cldf-metadata.json')
        if not md.exists():
            copy(Path(__file__).parent / 'cldf-metadata.json', md)
        self.wl = Wordlist.from_metadata(md)

        self.objects = {}
        self._obj_index = {}
        for cls in [
            self.dataset.lexeme_class,
            self.dataset.language_class,
            self.dataset.concept_class,
            self.dataset.cognate_class,
        ]:
            self.objects[cls.__cldf_table__()] = []
            self._obj_index[cls.__cldf_table__()] = set()

            cols = set(
                col.header for col in self.wl[cls.__cldf_table__()].tableSchema.columns)
            for field in cls.fieldnames():
                if field not in cols:
                    self.wl[cls.__cldf_table__()].tableSchema.columns.append(
                        Column(name=field, datatype="string"))

    def validate(self, log=None):
        return self.wl.validate(log)

    def __getitem__(self, type_):
        return self.wl[type_]

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        for table in ['FormTable', 'CognateTable', 'LanguageTable', 'ParameterTable']:
            self.objects.setdefault(table, [])
        self.write(**self.objects)

    def add_sources(self, *args):
        self.wl.sources.add(*args)

    def lexeme_id(self):
        self._count += 1
        return '{0}'.format(self._count)

    def cognate_id(self):
        self._cognate_count += 1
        return self._cognate_count

    def add_lexemes(self, **kw):
        """
        :return: list of dicts corresponding to newly created Lexemes
        """
        lexemes = []
        for i, form in enumerate(self.dataset.split_forms(kw, kw['Value'])):
            kw_ = kw.copy()
            if form:
                if form != kw_['Value']:
                    self.dataset.log.debug(
                        'iter_forms split: "{0}" -> "{1}"'.format(kw_['Value'], form))
                _form = self.dataset.clean_form(kw_, form)
                if _form != form:
                    self.dataset.log.debug(
                        'clean_form changed: "{0}" -> "{1}"'.format(form, _form))
                form = _form.strip()
                if form:
                    kw_.setdefault('Segments', self.dataset._segment(kw_, form) or [])
                    kw_.update(ID=self.lexeme_id(), Form=form)
                    lexemes.append(self._add_object(self.dataset.lexeme_class, **kw_))

        return lexemes

    def _add_object(self, cls, **kw):
        # Instantiating an object will trigger potential validators:
        d = attr.asdict(cls(**kw))
        for key in ['ID', 'Language_ID', 'Parameter_ID', 'Cognateset_ID']:
            # sluggify identifiers:
            if d.get(key):
                d[key] = slug('{0}'.format(d[key]))
        if 'ID' not in d or d['ID'] not in self._obj_index[cls.__cldf_table__()]:
            if 'ID' in d:
                self._obj_index[cls.__cldf_table__()].add(d['ID'])
            self.objects[cls.__cldf_table__()].append(d)
        return d

    def add_cognate(self, lexeme=None, **kw):
        if lexeme:
            kw.setdefault('Word_ID', lexeme['ID'])
            kw.setdefault('Form', lexeme['Form'])
        kw.setdefault('Index', self.cognate_id())
        return self._add_object(self.dataset.cognate_class, **kw)

    def add_language(self, **kw):
        return self._add_object(self.dataset.language_class, **kw)

    def add_concept(self, **kw):
        return self._add_object(self.dataset.concept_class, **kw)

    def align_cognates(self,
                       alm=None,
                       cognates=None,
                       column='Segments',
                       method='library'):
        from pylexibank.lingpy_util import iter_alignments

        iter_alignments(
            alm or self,
            cognates or self.objects['CognateTable'],
            column=column,
            method=method)

    def write(self, **kw):
        self.wl.properties.update(self.dataset.metadata.common_props)
        self.wl.properties['rdf:ID'] = self.dataset.id
        self.wl.tablegroup.notes.append({
            'dc:title': 'environment',
            'properties': {
                'concepticon_version': self.dataset.concepticon.version,
                'glottolog_version': self.dataset.glottolog.version,
            }
        })
        self.wl.write(**kw)



# coding: utf8
from __future__ import unicode_literals, print_function, division
from importlib import import_module

from clldutils.path import Path

for p in Path(__file__).parent.iterdir():
    if (p.is_file() and p.stem != '__init__' and p.name[0] != '.') \
            or (p.is_dir() and p.joinpath('__init__.py').exists()):
        import_module('{0}.{1}'.format(__name__, p.stem))



# coding=utf-8
from __future__ import unicode_literals, print_function
from collections import defaultdict, Counter
from copy import deepcopy

from clldutils.misc import slug
try:
    from lingpy.sequence.sound_classes import clean_string, tokens2class
    LINGPY = True
except ImportError:  # pragma: no cover
    LINGPY = False

    def clean_string(seq, *args, **kw):
        return [seq]

    def tokens2class(seq, *args, **kw):
        return []

import lingpy as lp
from pyclpa.base import get_clpa, Unknown, Sound
import attr

from pylexibank.dataset import Cognate
from pylexibank.util import get_url

clpa = get_clpa()


def getEvoBibAsBibtex(*keys, **kw):
    """Download bibtex format and parse it from EvoBib"""
    res = []
    for key in keys:
        bib = get_url(
            "http://bibliography.lingpy.org/raw.php?key=" + key,
            log=kw.get('log')).text
        try:
            res.append('@' + bib.split('@')[1].split('</pre>')[0])
        except IndexError:
            res.append('@misc{' + key + ',\nNote={missing source}\n\n}')
    return '\n\n'.join(res)


def wordlist2cognates(wordlist, source, expert='expert', ref='cogid'):
    """Turn a wordlist into a cognate set list, using the cldf parameters."""
    for k in wordlist:
        yield dict(
            Word_ID=wordlist[k, 'lid'],
            Index=k,
            Form=wordlist[k, 'ipa'],
            Cognateset_ID='{0}-{1}'.format(
                slug(wordlist[k, 'concept']), wordlist[k, ref]),
            Cognate_detection_method=expert,
            Cognate_source=source)


@attr.s
class TranscriptionAnalysis(object):
    # map segments to frequency
    segments = attr.ib(default=attr.Factory(Counter))
    # aggregate segments which are invalid for lingpy
    lingpy_errors = attr.ib(default=attr.Factory(set))
    # aggregate segments which are invalid for clpa
    clpa_errors = attr.ib(default=attr.Factory(set))
    # map clpa-replaceable segments to their replacements
    replacements = attr.ib(default=defaultdict(set))
    # count number of errors
    general_errors = attr.ib(default=0)


class InvalidString(Exception):
    pass


def test_sequence(segments, analysis=None, model='dolgo', **keywords):
    """
    Test a sequence for compatibility with CLPA and LingPy.

    :param analysis: Pass a `TranscriptionAnalysis` instance for cumulative reporting.
    """
    analysis = analysis or TranscriptionAnalysis()

    # clean the string at first, we only take the first item, ignore the rest
    try:
        if not isinstance(segments, list):
            segments = clean_string(segments, **keywords)[0].split(' ')
        lingpy_analysis = [
            x if y != '0' else '?' for x, y in
            zip(segments, tokens2class(segments, model))]
        clpa_analysis = clpa(segments)
        for l_, c in zip(lingpy_analysis, clpa_analysis):
            if l_ == '?' or isinstance(c, Unknown):
                analysis.general_errors += 1
    except (ValueError, IndexError, AttributeError, TypeError, AssertionError):
        raise InvalidString()

    for a, b, c in zip(segments, lingpy_analysis, clpa_analysis):
        if a[0] in clpa.accents:
            a = a[1:]
        analysis.segments.update([a])
        if b == '?':
            analysis.lingpy_errors.add(a)
        if isinstance(c, Unknown):
            analysis.clpa_errors.add(a)
        if isinstance(c, Sound) and c.converted:
            analysis.replacements[c.origin].add(c.clpa.CLPA)

    return segments, lingpy_analysis, ['%s' % x for x in clpa_analysis], analysis


def segmentize(string, clean=lambda s: s, **kw):
    """
    Write a detailed transcription-report for a CLDF dataset in LexiBank.
    """
    try:
        return test_sequence(clean(string), segmentized=False, **kw)[0]
    except InvalidString:
        pass


def test_sequences(dataset,
                   lid_getter,
                   model='dolgo'):
    """
    Write a detailed transcription-report for a CLDF dataset in LexiBank.
    """
    analyses, bad_words, invalid_words = {}, [], []

    for i, row in enumerate(dataset['FormTable']):
        analysis = analyses.setdefault(lid_getter(row), TranscriptionAnalysis())
        try:
            segments, _lpa, _clpa, _analysis = test_sequence(
                row['Segments'], analysis=analysis, model=model)
            if '?' in _lpa or '\ufffd' in _clpa:
                bad_words.append(row)
        except InvalidString:
            invalid_words.append(row)

    return analyses, bad_words, invalid_words


def _cldf2wld(dataset):
    """Make lingpy-compatible dictinary out of cldf main data."""
    header = [f for f in dataset.dataset.lexeme_class.fieldnames() if f != 'ID']
    D = {0: ['lid'] + [h.lower() for h in header]}
    for idx, row in enumerate(dataset.objects['FormTable']):
        row = deepcopy(row)
        row['Segments'] = ' '.join(row['Segments'])
        D[idx + 1] = [row['ID']] + [row[h] for h in header]
    return D


def _cldf2lexstat(
        dataset,
        segments='segments',
        transcription='value',
        row='parameter_name',
        col='language_name'):
    """Read LexStat object from cldf dataset."""
    D = _cldf2wld(dataset)
    return lp.LexStat(D, segments=segments, transcription=transcription, row=row, col=col)


def _cldf2wordlist(dataset, row='parameter_id', col='language_id'):
    """Read worldist object from cldf dataset."""
    return lp.Wordlist(_cldf2wld(dataset), row=row, col=col)


def iter_cognates(dataset, column='Segments', method='turchin', threshold=0.5, **kw):
    """
    Compute cognates automatically for a given dataset.
    """
    if method == 'turchin':
        for row in dataset.objects['FormTable']:
            sounds = ''.join(tokens2class(row[column], 'dolgo'))
            if sounds.startswith('V'):
                sounds = 'H' + sounds
            sounds = '-'.join([s for s in sounds if s != 'V'][:2])
            cogid = slug(row.get('Parameter_name') or row['Parameter_ID']) + '-' + sounds
            if '0' not in sounds:
                yield dict(
                    Word_ID=row['ID'],
                    Form=row['Value'],
                    Cognateset_ID=cogid,
                    Cognate_detection_method='CMM')

    if method in ['sca', 'lexstat']:
        lex = _cldf2lexstat(dataset)
        if method == 'lexstat':
            lex.get_scorer(**kw)
        lex.cluster(method=method, threshold=threshold, ref='cogid')
        for k in lex:
            yield Cognate(
                Word_ID=lex[k, 'lid'],
                Form=lex[k, 'value'],
                Cognateset_ID=lex[k, 'cogid'],
                Cognate_detection_method=method + '-t{0:.2f}'.format(threshold))


def iter_alignments(dataset, cognate_sets, column='Segments', method='library'):
    """
    Function computes automatic alignments and writes them to file.
    """
    if not isinstance(dataset, lp.basic.parser.QLCParser):
        wordlist = _cldf2wordlist(dataset)
        cognates = {r['Word_ID']: r for r in cognate_sets}
        wordlist.add_entries(
            'cogid',
            'lid',
            lambda x: cognates[x]['Cognateset_ID'] if x in cognates else '')
        for i, k in enumerate(wordlist):
            if not wordlist[k, 'cogid']:
                wordlist[k][wordlist.header['cogid']] = 'empty-%s' % i
        alm = lp.Alignments(
            wordlist,
            ref='cogid',
            row='parameter_id',
            col='language_id',
            segments=column.lower())
        alm.align(method=method)
        for k in alm:
            if alm[k, 'lid'] in cognates:
                cognate = cognates[alm[k, 'lid']]
                cognate['Alignment'] = alm[k, 'alignment'].split(' ')
                cognate['Alignment_method'] = method
    else:
        alm = lp.Alignments(dataset, ref='cogid')
        alm.align(method=method)

        for cognate in cognate_sets:
            idx = cognate['Index']
            if idx is None:
                idx = int(cognate['Word_ID'].split('-')[-1])
            cognate['Alignment'] = alm[idx, 'alignment'].split(' ')
            cognate['Alignment_method'] = 'SCA-' + method


def lingpy_subset(path, header, errors=2):
    try:
        wl = lp.get_wordlist(path, col='language_name', row='parameter_name')
    except ValueError:
        return []
    data = []

    if 'segments' not in wl.header:
        return []
    for taxon in wl.cols:
        error_count = 0
        idxs = wl.get_list(col=taxon, flat=True)
        goodlist = []
        for idx, segments in [(idx, wl[idx, 'segments']) for idx in idxs]:
            if wl[idx, 'language_id'] and wl[idx, 'parameter_id']:
                cv = lp.tokens2class(segments.split(), 'cv')
                if '0' in cv:
                    error_count += 1
                else:
                    l_ = sum(1 for x in cv if x != 'T')
                    if l_:
                        goodlist += [(idx, l_)]
                if error_count > errors:
                    goodlist = []
                    break
        for idx, l in goodlist:
            data.append([wl[idx, h] for h in header] + ['{0}'.format(l)])
    return data



# coding: utf8
from __future__ import unicode_literals, print_function, division
from zipfile import ZipFile

from clldutils.dsv import reader
from clldutils.path import TemporaryDirectory, write_text
from pycldf.dataset import MD_SUFFIX

from pylexibank.dataset import Dataset
from pylexibank.util import pb


class CLLD(Dataset):
    __cldf_url__ = None

    def url(self, path='/'):
        return "http://{0}.clld.org{1}".format(self.id, path)

    def cmd_download(self, **kw):
        self.raw.download(self.__cldf_url__, '{0}.zip'.format(self.id), log=self.log)
        self.raw.download(
            self.url(path="/resourcemap.json?rsc=language"),
            'languages.json',
            log=self.log)

    def add_sources(self, ds):
        archive = ZipFile(self.raw.joinpath('{0}.zip'.format(self.id)).as_posix())
        for name in archive.namelist():
            if name.endswith('.bib'):
                ds.add_sources(archive.read(name).decode('utf8'))

    def iteritems(self):
        archive = ZipFile(self.raw.joinpath('{0}.zip'.format(self.id)).as_posix())
        names = [name[:-len(MD_SUFFIX)]
                 for name in archive.namelist() if name.endswith(MD_SUFFIX)]
        with TemporaryDirectory() as tmp:
            for name in pb(names):
                write_text(tmp.joinpath('csv'), archive.read(name).decode('utf8'))
                for item in reader(tmp.joinpath('csv'), dicts=True):
                    item['Source'] = [item['Source']]
                    yield item



# coding: utf8
from __future__ import unicode_literals, print_function, division

from clldutils import jsonlib

from pylexibank.util import textdump
from pylexibank.commands.report import readme
from pylexibank.commands.report import transcription


def report(ds, **kw):
    """Create a README.md file listing the contents of a dataset

    lexibank report [DATASET_ID]
    """
    tr = jsonlib.load(ds.dir.joinpath('transcription.json'))
    textdump(
        transcription.report(tr, **kw),
        ds.dir.joinpath('TRANSCRIPTION.md'),
        log=kw.get('log'))
    res = readme.report(ds, tr, **kw)
    if res:
        textdump(res, ds.dir.joinpath('README.md'), log=kw.get('log'))



# coding: utf8
from __future__ import unicode_literals, print_function, division

from pyconcepticon import api

from pylexibank.util import Repos


class Concepticon(api.Concepticon, Repos):
    pass



# coding: utf8
"""Functionality to load a set of CLDF datasets into a sqlite db.

Notes:
- Only CLDF components will be loaded into the db.
- The names of the columns in the database are the names from the CSV files, not the
  preferred labels for the corresponding CLDF properties.
"""
from __future__ import unicode_literals, print_function, division
from collections import OrderedDict, defaultdict
import sqlite3
from contextlib import closing
from json import dumps

from six import text_type

import attr
from csvw.datatypes import DATATYPES
from clldutils.path import Path, remove
from clldutils.misc import nfilter

from pycldf.terms import term_uri
from pycldf.sources import Sources


from pylexibank.util import git_hash


def identity(s):
    return s


TYPE_MAP = {
    'string': ('TEXT', identity),
    'integer': ('INTEGER', identity),
    'boolean': ('INTEGER', lambda s: s if s is None else int(s)),
}
BIBTEX_FIELDS = [
    'address',  # Publisher's address
    'annote',  # An annotation for annotated bibliography styles (not typical)
    'author',  # The name(s) of the author(s) (separated by and)
    'booktitle',  # The title of the book, if only part of it is being cited
    'chapter',  # The chapter number
    'crossref',  # The key of the cross-referenced entry
    'edition',  # The edition of a book, long form (such as "First" or "Second")
    'editor',  # The name(s) of the editor(s)
    'eprint',  # A specification of electronic publication, preprint or technical report
    'howpublished',  # How it was published, if the publishing method is nonstandard
    'institution',  # institution involved in the publishing,not necessarily the publisher
    'journal',  # The journal or magazine the work was published in
    'key',  # A hidden field used for specifying or overriding the orderalphabetical order
    'month',  # The month of publication (or, if unpublished, the month of creation)
    'note',  # Miscellaneous extra information
    'number',  # The "(issue) number" of a journal, magazine, or tech-report
    'organization',  # The conference sponsor
    'pages',  # Page numbers, separated either by commas or double-hyphens.
    'publisher',  # The publisher's name
    'school',  # The school where the thesis was written
    'series',  # The series of books the book was published in
    'title',  # The title of the work
    'type',  # The field overriding the default type of publication
    'url',  # The WWW address
    'volume',  # The volume of a journal or multi-volume book
    'year',
]


def insert(db, table, keys, *rows):
    if rows:
        if isinstance(keys, text_type):
            keys = [k.strip() for k in keys.split(',')]
        db.executemany(
            "INSERT INTO {0} ({1}) VALUES ({2})".format(
                table, ','.join(keys), ','.join(['?' for _ in keys])),
            rows)


def quoted(*names):
    return ','.join('`{0}`'.format(name) for name in names)


@attr.s
class ColSpec(object):
    """
    A `ColSpec` captures sufficient information about a `Column` for the DB schema.
    """
    name = attr.ib()
    csvw_type = attr.ib(default='string', convert=lambda s: s if s else 'string')
    separator = attr.ib(default=None)
    primary_key = attr.ib(default=None)
    db_type = attr.ib(default=None)
    convert = attr.ib(default=None)

    def __attrs_post_init__(self):
        if self.csvw_type in TYPE_MAP:
            self.db_type, self.convert = TYPE_MAP[self.csvw_type]
        else:
            self.db_type = 'TEXT'
            self.convert = DATATYPES[self.csvw_type].to_string

    @property
    def sql(self):
        return '`{0.name}` {0.db_type}'.format(self)


@attr.s
class TableSpec(object):
    """
    A `TableSpec` captures sufficient information about a `Table` for the DB schema.
    """
    name = attr.ib()
    columns = attr.ib(default=attr.Factory(list))
    foreign_keys = attr.ib(default=attr.Factory(list))
    consumes = attr.ib(default=None)
    primary_key = attr.ib(default=None)

    @property
    def sql(self):
        clauses = [col.sql for col in self.columns]
        clauses.append('`dataset_ID` TEXT NOT NULL')
        if self.primary_key:
            clauses.append('PRIMARY KEY(`dataset_ID`, `{0}`)'.format(self.primary_key))
        clauses.append('FOREIGN KEY(`dataset_ID`) REFERENCES dataset(`ID`)')
        for fk, ref, refcols in self.foreign_keys:
            clauses.append('FOREIGN KEY({0}) REFERENCES {1}({2})'.format(
                quoted(*fk), ref, quoted(*refcols)))
        return "CREATE TABLE {0} (\n    {1}\n)".format(self.name, ',\n    '.join(clauses))


def schema(ds):
    """
    Convert the table and column descriptions of a `Dataset` into specifications for the
    DB schema.

    :param ds:
    :return: A pair (tables, reference_tables).
    """
    tables, ref_tables = {}, {}
    table_lookup = {t.url.string: t for t in ds.tables if ds.get_tabletype(t)}
    for table in table_lookup.values():
        spec = TableSpec(ds.get_tabletype(table))
        spec.primary_key = [
            c for c in table.tableSchema.columns if
            c.propertyUrl and c.propertyUrl.uri == term_uri('id')][0].name
        for c in table.tableSchema.columns:
            if c.propertyUrl and c.propertyUrl.uri == term_uri('source'):
                # A column referencing sources is replaced by an association table.
                otype = ds.get_tabletype(table).replace('Table', '')
                ref_tables[ds.get_tabletype(table)] = TableSpec(
                    '{0}Source'.format(otype),
                    [ColSpec(otype + '_ID'), ColSpec('Source_ID'), ColSpec('Context')],
                    [
                        (
                            ['dataset_ID', otype + '_ID'],
                            ds.get_tabletype(table),
                            ['dataset_ID', spec.primary_key]),
                        (
                            ['dataset_ID', 'Source_ID'],
                            'SourceTable',
                            ['dataset_ID', 'ID']),
                    ],
                    c.name)
            else:
                spec.columns.append(ColSpec(
                    c.header,
                    c.datatype.base if c.datatype else c.datatype,
                    c.separator,
                    c.header == spec.primary_key))
        for fk in table.tableSchema.foreignKeys:
            if fk.reference.schemaReference:
                # We only support Foreign Key references between tables!
                continue  # pragma: no cover
            ref = table_lookup[fk.reference.resource.string]
            if ds.get_tabletype(ref):
                spec.foreign_keys.append((
                    tuple(['dataset_ID'] + sorted(fk.columnReference)),
                    ds.get_tabletype(table_lookup[fk.reference.resource.string]),
                    tuple(['dataset_ID'] + sorted(fk.reference.columnReference))))
        tables[spec.name] = spec

    # must determine the order in which tables must be created!
    ordered = OrderedDict()
    i = 0
    #
    # We loop through the tables repeatedly, and whenever we find one, which has all
    # referenced tables already in ordered, we move it from tables to ordered.
    #
    while tables and i < 100:
        i += 1
        for table in list(tables.keys()):
            if all(ref[1] in ordered for ref in tables[table].foreign_keys):
                # All referenced tables are already created.
                ordered[table] = tables.pop(table)
                break
    if tables:  # pragma: no cover
        raise ValueError('there seem to be cyclic dependencies between the tables')

    return list(ordered.values()), ref_tables


class Database(object):
    def __init__(self, fname):
        """
        A `Database` instance is initialized with a file path.

        :param fname: Path to a file in the file system where the db is to be stored.
        """
        self.fname = Path(fname)

    def drop(self):
        if self.fname.exists():
            remove(self.fname)

    def connection(self):
        return closing(sqlite3.connect(self.fname.as_posix()))

    def create(self, force=False, exists_ok=False):
        """
        Creates a db file with the core schema.

        :param force: If `True` an existing db file will be overwritten.
        """
        if self.fname and self.fname.exists():
            if force:
                self.drop()
            elif exists_ok:
                return
            else:
                raise ValueError('db file already exists, use force=True to overwrite')
        with self.connection() as db:
            db.execute(
                """\
CREATE TABLE dataset (
    ID TEXT PRIMARY KEY NOT NULL,
    name TEXT,
    version TEXT,
    metadata_json TEXT
)""")
            db.execute("""\
CREATE TABLE datasetmeta (
    dataset_ID TEXT ,
    key TEXT,
    value TEXT,
    PRIMARY KEY (dataset_ID, key),
    FOREIGN KEY(dataset_ID) REFERENCES dataset(ID)
)""")
            db.execute("""\
CREATE TABLE SourceTable (
    dataset_ID TEXT ,
    ID TEXT ,
    bibtex_type TEXT,
    {0}
    extra TEXT,
    PRIMARY KEY (dataset_ID, ID),
    FOREIGN KEY(dataset_ID) REFERENCES dataset(ID)
)""".format('\n    '.join('`{0}` TEXT,'.format(f) for f in BIBTEX_FIELDS)))

    def fetchone(self, sql, conn=None, verbose=False):
        return self._fetch(sql, 'fetchone', conn, verbose=verbose)

    def fetchall(self, sql, conn=None, verbose=False):
        return self._fetch(sql, 'fetchall', conn, verbose=verbose)

    def _fetch(self, sql, method, conn, verbose=False):
        sql = self.sql.get(sql, sql)

        def _do(conn, sql, method):
            cu = conn.cursor()
            if verbose:
                print(sql)
            cu.execute(sql)
            return getattr(cu, method)()

        if not conn:
            with self.connection() as conn:
                return _do(conn, sql, method)
        else:
            return _do(conn, sql, method)

    def unload(self, dataset_id):
        dataset_id = getattr(dataset_id, 'id', dataset_id)
        with self.connection() as db:
            for row in db.execute("SELECT name FROM sqlite_master WHERE type='table'"):
                table = row[0]
                if table != 'dataset':
                    db.execute(
                        "DELETE FROM {0} WHERE dataset_ID = ?".format(table),
                        (dataset_id,))
            db.execute("DELETE FROM dataset WHERE ID = ?", (dataset_id,))
            db.commit()

    def _create_table_if_not_exists(self, table):
        if table.name in [r[0] for r in self.fetchall(
                "SELECT name FROM sqlite_master WHERE type='table'")]:
            return False

        with self.connection() as conn:
            conn.execute(table.sql)
        return True

    def load(self, ds):
        """
        Load a CLDF dataset into the database.

        :param dataset:
        :return:
        """
        try:
            self.fetchone('select ID from dataset')
        except sqlite3.OperationalError:
            self.create(force=True)
        self.unload(ds)
        dataset = ds.cldf.wl
        tables, ref_tables = schema(dataset)

        # update the DB schema:
        for t in tables:
            if self._create_table_if_not_exists(t):
                continue
            db_cols = {r[1]: r[2] for r in self.fetchall(
                "PRAGMA table_info({0})".format(t.name))}
            for col in t.columns:
                if col.name not in db_cols:
                    with self.connection() as conn:
                        conn.execute(
                            "ALTER TABLE {0} ADD COLUMN `{1.name}` {1.db_type}".format(
                                t.name, col))
                else:
                    if db_cols[col.name] != col.db_type:
                        raise ValueError(
                            'column {0}:{1} {2} redefined with new type {3}'.format(
                                t.name, col.name, db_cols[col.name], col.db_type))

        for t in ref_tables.values():
            self._create_table_if_not_exists(t)

        # then load the data:
        with self.connection() as db:
            db.execute('PRAGMA foreign_keys = ON;')
            insert(
                db,
                'dataset',
                'ID,name,version,metadata_json',
                (
                    ds.id,
                    '{0}'.format(dataset),
                    git_hash(ds.dir),
                    dumps(dataset.metadata_dict)))
            insert(
                db,
                'datasetmeta',
                'dataset_ID,key,value',
                *[(ds.id, k, '{0}'.format(v)) for k, v in dataset.properties.items()])

            # load sources:
            rows = []
            for src in dataset.sources.items():
                values = [ds.id, src.id, src.genre] + [src.get(k) for k in BIBTEX_FIELDS]
                values.append(
                    dumps({k: v for k, v in src.items() if k not in BIBTEX_FIELDS}))
                rows.append(tuple(values))
            insert(
                db,
                'SourceTable',
                ['dataset_ID', 'ID', 'bibtex_type'] + BIBTEX_FIELDS + ['extra'],
                *rows)

            # For regular tables, we extract and keep references to sources.
            refs = defaultdict(list)

            for t in tables:
                cols = {col.name: col for col in t.columns}
                ref_table = ref_tables.get(t.name)
                rows, keys = [], []
                for row in dataset[t.name]:
                    keys, values = ['dataset_ID'], [ds.id]
                    for k, v in row.items():
                        if ref_table and k == ref_table.consumes:
                            refs[ref_table.name].append((row[t.primary_key], v))
                        else:
                            col = cols[k]
                            if isinstance(v, list):
                                v = (col.separator or ';').join(
                                    nfilter(col.convert(vv) for vv in v))
                            else:
                                v = col.convert(v)
                            keys.append("`{0}`".format(k))
                            values.append(v)
                    rows.append(tuple(values))
                insert(db, t.name, keys, *rows)

            # Now insert the references, i.e. the associations with sources:
            for tname, items in refs.items():
                rows = []
                for oid, sources in items:
                    for source in sources:
                        sid, context = Sources.parse(source)
                        rows.append([ds.id, oid, sid, context])
                oid_col = '{0}_ID'.format(tname.replace('Source', ''))
                insert(db, tname, ['dataset_ID', oid_col, 'Source_ID', 'Context'], *rows)
            db.commit()

    def load_concepticon_data(self, concepticon):
        conceptsets = []
        for csid in self.fetchall("SELECT distinct conceptset FROM parametertable"):
            cs = concepticon.conceptsets.get(csid[0])
            if cs:
                conceptsets.append((cs.gloss, cs.id))

        with self.connection() as db:
            db.executemany(
                "UPDATE parametertable SET concepticon_gloss = ? WHERE conceptset = ?",
                conceptsets)
            db.commit()

    def load_glottolog_data(self, glottolog):
        langs = []
        languoids = {l.id: l for l in glottolog.languoids()}
        for gc in self.fetchall("SELECT distinct glottocode FROM languagetable"):
            lang = languoids.get(gc[0])
            if lang:
                langs.append((
                    lang.lineage[0][0] if lang.lineage else lang.name,
                    lang.macroareas[0].value if lang.macroareas else None,
                    lang.id))

        with self.connection() as db:
            db.executemany(
                "UPDATE languagetable "
                "SET family = ?, macroarea = ? "
                "WHERE glottocode = ?",
                langs)
            db.commit()

    sql = {
        "conceptsets_by_dataset":
            "SELECT ds.id, count(distinct p.conceptset) "
            "FROM dataset as ds, parametertable as p "
            "WHERE ds.id = p.dataset_id GROUP BY ds.id",
        "families_by_dataset":
            "SELECT ds.id, count(distinct l.family) "
            "FROM dataset as ds, languagetable as l "
            "WHERE ds.id = l.dataset_id GROUP BY ds.id",
        "macroareas_by_dataset":
            "SELECT ds.id, group_concat(distinct l.macroarea) "
            "FROM dataset as ds, languagetable as l "
            "WHERE ds.id = l.dataset_id GROUP BY ds.id",
        "glottocodes_by_dataset":
            "SELECT ds.id, count(distinct l.glottocode) "
            "FROM dataset as ds, languagetable as l "
            "WHERE ds.id = l.dataset_id GROUP BY ds.id",
        "mapped_lexemes_by_dataset":
            "SELECT ds.id, count(f.ID) "
            "FROM dataset as ds, formtable as f, languagetable as l, parametertable as p "
            "WHERE ds.id = f.dataset_id and f.Language_ID = l.ID and "
            "f.Parameter_ID = p.ID and l.glottocode is not null and "
            "p.conceptset is not null "
            "GROUP BY ds.id",
        "lexemes_by_dataset":
            "SELECT ds.id, count(f.ID) FROM dataset as ds, formtable as f "
            "WHERE ds.id = f.dataset_id GROUP BY ds.id",
    }



"""
Main command line interface of the pylexibank package.

Like programs such as git, this cli splits its functionality into sub-commands
(see e.g. https://docs.python.org/2/library/argparse.html#sub-commands).
The rationale behind this is that while a lot of different tasks may be
triggered using this cli, most of them require common configuration.

The basic invocation looks like

    lexibank [OPTIONS] <command> [args]

"""
from __future__ import unicode_literals, division, print_function
import sys
import os
import argparse
import readline
import glob
import pkg_resources

from termcolor import colored
from appdirs import user_config_dir
from six.moves import input
from clldutils.inifile import INI
from clldutils.clilib import ArgumentParserWithLogging
from clldutils.path import Path

import pylexibank
from pylexibank.glottolog import Glottolog
from pylexibank.concepticon import Concepticon
import pylexibank.commands
assert pylexibank.commands

REPOS = [
    ('glottolog', 'clld/glottolog'),
    ('concepticon', 'clld/concepticon-data'),
]


# We want to provide tab-completion when the user is asked to provide local paths to
# repository clones.
def complete_dir(text, state):  # pragma: no cover
    if os.path.isdir(text) and not text.endswith(os.sep):
        text += os.sep
    return ([p for p in glob.glob(text + '*') if os.path.isdir(p)] + [None])[state]


readline.parse_and_bind("tab: complete")
readline.set_completer_delims('\t')
readline.set_completer(complete_dir)


def get_path(src):  # pragma: no cover
    """
    Prompts the user to input a local path.

    :param src: github repository name
    :return: Absolute local path
    """
    res = None
    while not res:
        if res is False:
            print(colored('You must provide a path to an existing directory!', 'red'))
        print('You need a local clone or release of (a fork of) '
              'https://github.com/{0}'.format(src))
        res = input(colored('Local path to {0}: '.format(src), 'green', attrs=['blink']))
        if res and Path(res).exists():
            return Path(res).resolve()
        res = False


def configure():  # pragma: no cover
    """
    Configure lexibank.

    :return: a pair (config, logger)
    """
    cfgpath = Path(user_config_dir(pylexibank.__name__)).joinpath('config.ini')
    if not cfgpath.exists():
        print("""
{0}

You seem to be running lexibank for the first time.
Your system configuration will now be written to a config file to be used
whenever lexibank is run lateron.
""".format(
            colored('Welcome to lexibank!', 'blue', attrs=['bold', 'reverse'])))
        if not cfgpath.parent.exists():
            cfgpath.parent.mkdir(parents=True)
        cfg = INI()
        cfg['paths'] = {k: get_path(src) for k, src in REPOS}
        cfg.write(cfgpath)
        print("""
Configuration has been written to:
{0}
You may edit this file to adapt to changes in your system or to reconfigure settings
such as the logging level.""".format(cfgpath.resolve()))
    else:
        cfg = INI.from_file(cfgpath)

    datasets = []
    for ep in pkg_resources.iter_entry_points('lexibank.dataset'):
        ds_class = ep.load()
        datasets.append(ds_class(
            glottolog=Glottolog(cfg['paths']['glottolog']),
            concepticon=Concepticon(cfg['paths']['concepticon'])))

    return cfg, sorted(datasets, key=lambda d: d.id)


def main():  # pragma: no cover
    cfg, datasets = configure()
    parser = ArgumentParserWithLogging(pylexibank.__name__)
    parser.add_argument('--cfg', help=argparse.SUPPRESS, default=cfg)
    parser.add_argument('--datasets', help=argparse.SUPPRESS, default=datasets)
    parser.add_argument(
        '--db',
        help=argparse.SUPPRESS,
        default=os.path.join(os.getcwd(), 'lexibank.sqlite'))
    sys.exit(parser.main())



# coding: utf8
from __future__ import unicode_literals, print_function, division
import logging
from datetime import datetime
import re

import attr
import six
import git
from clldutils.dsv import reader
from clldutils.text import split_text_with_context, strip_brackets
from clldutils.misc import lazyproperty
from clldutils.loglib import Logging
from clldutils.path import remove, rmtree, write_text
from clldutils import licenses
from clldutils import jsonlib
from pyglottolog.languoids import Glottocode

from lingpy.settings import rc
from segments.tokenizer import Tokenizer

import pylexibank
from pylexibank.util import get_variety_id, DataDir
from pylexibank import cldf

NOOP = -1


def non_empty(_, attribute, value):
    if not value:
        raise ValueError('{0} must be non-empty'.format(attribute))


class FieldnamesMixin(object):
    @classmethod
    def fieldnames(cls):
        return [f.name for f in attr.fields(cls)]


@attr.s(hash=False)
class Language(FieldnamesMixin):
    ID = attr.ib(default='', hash=True)
    name = attr.ib(default=None)
    iso = attr.ib(default=None)
    glottocode = attr.ib(default=None)
    glottolog_name = attr.ib(default=None)
    macroarea = attr.ib(default=None)
    family = attr.ib(default=None)

    @classmethod
    def __cldf_table__(cls):
        return 'LanguageTable'


@attr.s(hash=False)
class Concept(FieldnamesMixin):
    ID = attr.ib(default='', hash=True)
    gloss = attr.ib(default='')
    conceptset = attr.ib(default='')
    concepticon_gloss = attr.ib(default='')

    @classmethod
    def __cldf_table__(cls):
        return 'ParameterTable'


@attr.s
class Lexeme(FieldnamesMixin):
    """
    Raw lexical data item as it can be pulled out of the original datasets.

    This is the basis for creating rows in CLDF representations of the data by
    - splitting the lexical item into forms
    - cleaning the forms
    - potentially tokenizing the form
    """
    ID = attr.ib()
    Form = attr.ib()
    Value = attr.ib(validator=non_empty)  # the lexical item
    Language_ID = attr.ib(validator=non_empty)
    Parameter_ID = attr.ib(validator=non_empty)
    Local_ID = attr.ib(default=None)
    Segments = attr.ib(
        default=attr.Factory(list),
        validator=attr.validators.instance_of(list))
    Source = attr.ib(
        default=attr.Factory(list),
        validator=attr.validators.instance_of(list),
        convert=lambda v: [v] if isinstance(v, six.string_types) else v)
    Comment = attr.ib(default=None)
    Cognacy = attr.ib(default=None)
    Loan = attr.ib(
        default=None,
        validator=attr.validators.optional(attr.validators.instance_of(bool)))

    @classmethod
    def __cldf_table__(cls):
        return 'FormTable'


@attr.s
class Cognate(FieldnamesMixin):
    Word_ID = attr.ib(default=None)  # The lexibank-wide unique word ID.
    Index = attr.ib(default=None)  # integer index used in lingpy wordlists.
    Form = attr.ib(default=None)
    Cognateset_ID = attr.ib(default=None)
    Doubt = attr.ib(
        default=False, convert=lambda v: v if isinstance(v, bool) else eval(v))
    Cognate_detection_method = attr.ib(default='expert')
    Cognate_source = attr.ib(default=None)
    Alignment = attr.ib(
        default=None, convert=lambda v: ' '.join(v) if isinstance(v, list) else v)
    Alignment_method = attr.ib(default=None)
    Alignment_source = attr.ib(default=None)

    @classmethod
    def __cldf_table__(cls):
        return 'CognateTable'


@attr.s
class Metadata(object):
    title = attr.ib(default=None)
    description = attr.ib(default=None)
    citation = attr.ib(default=None)
    license = attr.ib(default=None)
    url = attr.ib(default=None)
    aboutUrl = attr.ib(default=None)
    conceptlist = attr.ib(default=None)
    lingpy_schema = attr.ib(default=None)
    derived_from = attr.ib(default=None)
    related = attr.ib(default=None)
    source = attr.ib(default=None)

    @property
    def common_props(self):
        return {
            "dc:title": self.title,
            "dc:description": self.description,
            "dc:bibliographicCitation": self.citation,
            "dc:license": licenses.find(self.license) if self.license else None,
            "dc:identifier": self.url,
            "dc:format": "http://concepticon.clld.org/contributions/{0}".format(
                self.conceptlist) if self.conceptlist else None,
            "dc:isVersionOf": "http://lexibank.clld.org/contributions/{0}".format(
                self.derived_from) if self.derived_from else None,
            "dc:related": self.related,
            "aboutUrl": self.aboutUrl,
        }


class Dataset(object):
    """
    A lexibank dataset.

    This object provides access to a dataset's
    - language list as attribute `languages`
    - concept list as attribute `concepts`
    - concepticon concept-list ID as attribute `conceptlist`
    """
    dir = None  # Derived classes must provide an existing directory here!
    lexeme_class = Lexeme
    cognate_class = Cognate
    language_class = Language
    concept_class = Concept
    log = logging.getLogger(pylexibank.__name__)

    @lazyproperty
    def id(self):
        return self.dir.name

    @lazyproperty
    def metadata(self):
        return Metadata(**jsonlib.load(self.dir / 'metadata.json'))

    @property
    def stats(self):
        if self.dir.joinpath('README.json').exists():
            return jsonlib.load(self.dir / 'README.json')
        return {}

    def __init__(self, concepticon=None, glottolog=None):
        self.unmapped = Unmapped()
        self.dir = DataDir(self.dir)
        self.raw = DataDir(self.dir / 'raw')
        self.cldf_dir = self.dir / 'cldf'

        self.conceptlist = {}
        self.glottolog = glottolog
        self.concepticon = concepticon
        try:
            self.git_repo = git.Repo(str(self.dir))  # pragma: no cover
        except git.InvalidGitRepositoryError:
            self.git_repo = None

    def _iter_etc(self, what):
        path = self.dir / 'etc' / what
        return reader(path, dicts=True) if path.exists() else []

    @lazyproperty
    def github_repo(self):  # pragma: no cover
        try:
            match = re.search(
                'github\.com/(?P<org>[^/]+)/%s\.git' % re.escape(self.id),
                self.git_repo.remotes.origin.url)
            if match:
                return match.group('org') + '/' + self.id
        except AttributeError:
            pass

    @lazyproperty
    def sources(self):
        return list(self._iter_etc('sources.csv'))

    @lazyproperty
    def concepts(self):
        return list(self._iter_etc('concepts.csv'))

    @lazyproperty
    def languages(self):
        res = []
        for item in self._iter_etc('languages.csv'):
            if item.get('GLOTTOCODE', None) and not \
                    Glottocode.pattern.match(item['GLOTTOCODE']):  # pragma: no cover
                raise ValueError(
                    "Wrong glottocode for item {0}".format(item['GLOTTOCODE']))
            res.append(item)
        return res

    @lazyproperty
    def lexemes(self):
        res = {}
        for item in self._iter_etc('lexemes.csv'):
            res[item['LEXEME']] = item['REPLACEMENT']
        return res

    def debug(self):
        return Logging(self.log)

    # ---------------------------------------------------------------
    # workflow actions whih should be overwritten by derived classes:
    # ---------------------------------------------------------------
    def cmd_download(self, **kw):
        self._not_implemented('download')
        return NOOP

    def cmd_install(self, **kw):
        self._not_implemented('install')
        return NOOP

    # ---------------------------------------------------------------
    # handling of lexemes/forms/words
    # ---------------------------------------------------------------
    def clean_form(self, item, form):
        """
        Called when a row is added to a CLDF dataset.

        :param form:
        :return: None to skip the form, or the cleaned form as string.
        """
        form = strip_brackets(form)
        if form not in ['?']:
            return form

    def split_forms(self, item, value):
        value = self.lexemes.get(value, value)
        return [self.clean_form(item, form)
                for form in split_text_with_context(value, separators='/,;')]

    def get_tokenizer(self):
        profile = self.dir / 'etc' / 'orthography.tsv'
        if profile.exists():
            obj = Tokenizer(profile=str(profile))

            def _tokenizer(item, string, **kw):
                kw.setdefault("column", "IPA")
                kw.setdefault("separator", " _ ")
                return obj(string, **kw).split()
            return _tokenizer

    # ---------------------------------------------------------------
    # CLDF dataset access
    # ---------------------------------------------------------------
    @property
    def cldf(self):
        return cldf.Dataset(self)

    # ---------------------------------------------------------------
    def _download(self, **kw):
        if not self.raw.exists():
            self.raw.mkdir()

        self.cmd_download(**kw)
        write_text(
            self.raw / 'README.md',
            'Raw data downloaded {0}'.format(datetime.utcnow().isoformat()))

    def _install(self, **kw):
        if not self.cldf_dir.exists():
            self.cldf_dir.mkdir()

        self.unmapped.clear()

        if self.metadata.conceptlist:
            self.conceptlist = self.concepticon.conceptlists[self.metadata.conceptlist]
        rc(schema=self.metadata.lingpy_schema or 'ipa')
        if self.cmd_install(**kw) != NOOP:
            if kw.get('verbose'):
                self.unmapped.pprint()
            self.cldf.validate(kw['log'])

    def _clean(self, **kw):
        self.log.debug('removing CLDF directory %s' % self.cldf_dir)
        if self.cldf_dir.exists():
            for f in self.cldf_dir.iterdir():
                if f.is_file():
                    remove(f)
                else:
                    rmtree(f)

    def _not_implemented(self, method):
        self.log.warn('cmd_{0} not implemented for dataset {1}'.format(method, self.id))

    @lazyproperty
    def _tokenizer(self):
        return self.get_tokenizer()

    def _segment(self, item, string):
        if self._tokenizer:
            return self._tokenizer(item, string)

    def coverage(self, vars, glangs, c):  # pragma: no cover
        for row in self.cldf['FormTable']:
            try:
                cid = int(row['Parameter_ID'])
            except (ValueError, TypeError):
                continue
            vid = self.id + '-' + get_variety_id(row)
            c[cid].add(vid)
            vars[vid].add(cid)
            glangs[row['Language_ID']].add(cid)


class Unmapped(object):
    def __init__(self):
        self.languages = set()
        self.concepts = set()

    def clear(self):
        self.languages = set()
        self.concepts = set()

    def add_concept(self, **kw):
        kw['ID'] = kw.pop('id')
        self.concepts.add(Concept(**kw))

    def add_language(self, **kw):
        kw['ID'] = kw.pop('id')
        self.languages.add(Language(**kw))

    @staticmethod
    def quote(v):
        v = '{0}'.format(v or '')
        if ',' in v or len(v.split()) > 1:
            v = '"%s"' % v.replace('"', '""')
        return v

    def pprint(self):
        for objs, cls in [(self.languages, Language), (self.concepts, Concept)]:
            if objs:
                print('=== Unmapped %ss ===' % cls.__name__)
                print(','.join([a.name.upper() for a in attr.fields(cls)]))
                for row in sorted(map(attr.astuple, objs)):
                    print(','.join(map(self.quote, row)))



# *-* coding: utf-8 *-*
"""
Implements the readme file generator.
"""
from __future__ import division, unicode_literals, print_function
from collections import Counter, defaultdict

from pylexibank.util import get_badge, get_variety_id, jsondump


def build_status_badge(ds):
    if not ds.dir.joinpath('.travis.yml').exists():
        return ''
    try:
        return "[![Build Status](https://travis-ci.org/{0}.svg?branch=master)]" \
               "(https://travis-ci.org/{0})".format(ds.github_repo)
    except:
        return ''


def report(ds, tr_analysis, log=None, **kw):
    #
    # FIXME: write only summary into README.md
    # in case of multiple cldf datasets:
    # - separate lexemes.md and transcriptions.md
    #
    if not list(ds.cldf_dir.glob('*.csv')):
        return
    lines = [
        '# %s' % ds.metadata.title,
        '',
        'Cite the source dataset as',
        '',
        '> %s' % ds.metadata.citation,
        '']

    if ds.metadata.license:
        lines.extend([
            'This dataset is licensed under a %s license' % ds.metadata.license, ''])

    if ds.metadata.url:
        lines.extend(['Available online at %s' % ds.metadata.url, ''])

    if ds.metadata.related:
        lines.extend(['See also %s' % ds.metadata.related, ''])

    # add NOTES.md
    if ds.dir.joinpath('NOTES.md').exists():
        lines.extend(['## Notes', ''])
        lines.extend(ds.dir.joinpath('NOTES.md').read_text().split("\n"))
        lines.extend(['', ''])  # some blank lines

    trlines = []

    synonyms = defaultdict(Counter)
    totals = {
        'languages': Counter(),
        'concepts': Counter(),
        'sources': Counter(),
        'cognate_sets': Counter(),
        'lexemes': 0,
    }

    missing_source = []
    missing_lang = [L['NAME'] for L in ds.languages if not L['GLOTTOCODE']]
    missing_param = []
    missing_concept = [c for c in ds.concepts if not c['CONCEPTICON_ID']]

    for row in ds.cldf['FormTable']:
        if row['Source']:
            totals['sources'].update(['y'])
        else:
            missing_source.append(row)
        if row['Parameter_ID']:
            totals['concepts'].update([row['Parameter_ID']])
        else:
            missing_param.append(row)
        lid = get_variety_id(row)
        totals['languages'].update([lid])
        totals['lexemes'] += 1
        if lid and row['Parameter_ID']:
            synonyms[lid].update([row['Parameter_ID']])

    for row in ds.cldf['CognateTable']:
        totals['cognate_sets'].update([row['Cognateset_ID']])

    sindex = sum(
        [sum(list(counts.values())) / float(len(counts)) for counts in synonyms.values()])
    langs = set(synonyms.keys())
    if langs:
        sindex /= float(len(langs))
    else:
        sindex = 0
    totals['SI'] = sindex

    stats = tr_analysis['stats']
    lsegments = len(stats['segments'])
    llingpyerr = len(stats['lingpy_errors'])
    lclpaerr = len(stats['clpa_errors'])

    def ratio(prop):
        return sum(v for k, v in totals[prop].items() if k) / float(totals['lexemes'])

    badges = [
        build_status_badge(ds),
        get_badge(ratio('languages'), 'Glottolog'),
        get_badge(ratio('concepts'), 'Concepticon'),
        get_badge(ratio('sources'), 'Source'),
    ]
    if lsegments:
        badges.extend([
            get_badge((lsegments - llingpyerr) / lsegments, 'LingPy'),
            get_badge((lsegments - lclpaerr) / lsegments, 'CLPA'),
        ])
    lines.extend(['## Statistics', '\n', '\n'.join(badges), ''])
    stats_lines = [
        '- **Varieties:** {0:,}'.format(len(totals['languages'])),
        '- **Concepts:** {0:,}'.format(len(totals['concepts'])),
        '- **Lexemes:** {0:,}'.format(totals['lexemes']),
        '- **Synonymy:** %.2f' % (totals['SI']),
        '- **Cognacy:** {0:,} cognates in {1:,} cognate sets'.format(
            sum(v for k, v in totals['cognate_sets'].items() if v > 1),
            sum(1 for k, v in totals['cognate_sets'].items() if v > 1)),
        '- **Invalid lexemes:** {0:,}'.format(stats['invalid_words_count']),
        '- **Tokens:** {0:,}'.format(sum(stats['segments'].values())),
        '- **Segments:** {0:,} ({1} LingPy errors, {2} CLPA errors, {3} CLPA modified)'
        .format(lsegments, llingpyerr, lclpaerr, len(stats['replacements'])),
        '- **Inventory size (avg):** %.2f' % stats['inventory_size'],
    ]
    if log:
        log.info('\n'.join(['Summary for dataset {}'.format(ds.id)] + stats_lines))
    lines.extend(stats_lines)

    # improvements section
    if len(missing_lang) or len(missing_source) or len(missing_concept):
        lines.extend(['\n## Possible Improvements:\n', ])
        
        if len(missing_lang):
            lines.append("- Languages missing glottocodes: %d/%d (%.2f%%)" % (
                len(missing_lang),
                len(totals['languages']),
                (len(missing_lang) / len(totals['languages'])) * 100
            ))
            
        if len(missing_source):
            lines.append("- Entries missing sources: %d/%d (%.2f%%)" % (
                len(missing_source),
                totals['lexemes'],
                (len(missing_source) / totals['lexemes']) * 100
            ))

        if len(missing_concept):
            lines.append("- Entries missing concepts: %d/%d (%.2f%%)" % (
                    len(missing_param),
                    totals['lexemes'],
                    (len(missing_param) / totals['lexemes']) * 100
            ))
        lines.append("\n")

    jsondump(totals, ds.dir.joinpath('README.json'))
    return lines + trlines



# coding: utf8
from __future__ import unicode_literals, print_function, division

from pylexibank.dataset import Dataset


class QLC(Dataset):
    DSETS = []

    #def cmd_download(self, **kw):
    #    self.raw.download_and_unpack(
    #        "http://www.quanthistling.info/data/downloads/csv/data.zip",
    #        *[name for name in self.DSETS],
    #        **{'log': self.log})



# coding: utf8
from __future__ import unicode_literals, print_function, division
from time import time

from clldutils.path import Path
from clldutils.clilib import ParserError

from pylexibank.db import Database


def get_dataset(args, name=None):
    id_ = Path(name or args.args[0]).name
    for dataset in args.datasets:
        if dataset.id == id_:
            return dataset
    raise ParserError('invalid dataset spec')  # pragma: no cover


def with_dataset(args, func):
    for dataset in args.datasets:
        if dataset.id in args.args:
            s = time()
            if args.verbosity:
                print('processing %s ...' % dataset.id)
            func(get_dataset(args, dataset.id), **vars(args))
            if args.verbosity:
                print('... done %s [%.1f secs]' % (dataset.id, time() - s))


def _load(ds, **kw):
    db = Database(kw['db'])
    db.create(exists_ok=True)
    db.load(ds)
    db.load_concepticon_data(ds.concepticon)
    db.load_glottolog_data(ds.glottolog)


def _unload(ds, **kw):
    db = Database(kw['db'])
    db.create(exists_ok=True)
    db.unload(ds)



# coding: utf8
from __future__ import unicode_literals, print_function, division

from clldutils.markup import Table

from pylexibank.util import get_variety_id

TEMPLATE = """
# Detailed transcription record

## Segments

{0}

## Unsegmentable lexemes (up to 100 only)

{1}

## Words with invalid segments (up to 100 only)

{2}
"""

MARKDOWN_TEMPLATE = """
## Transcription Report

### General Statistics

* Number of Tokens: {tokens}
* Number of Segments: {segments}
* Invalid forms: {invalid}
* Inventory Size: {inventory_size:.2f}
* [Erroneous tokens](report.md#tokens): {general_errors}
* Erroneous words: {word_errors}
* Number of LingPy-Errors: {lingpy_errors}
* Number of CLPA-Errors: {clpa_errors}
* Bad words: {words_errors}
"""


def report(analysis, **kw):
    segments = Table('Segment', 'Occurrence', 'LingPy', 'CLPA')
    for a, b in sorted(
            analysis['stats']['segments'].items(), key=lambda x: (-x[1], x[0])):
        c, d = '✓', '✓'
        if a in analysis['stats']['clpa_errors']:
            c = '✓' if a not in analysis['stats']['lingpy_errors'] else '?'
            d = ', '.join(analysis['stats']['clpa_errors'][a]) \
                if a not in analysis['stats']['clpa_errors'] else '?'

        # escape pipe for markdown table if necessary
        a = a.replace('|', '&#124;')

        segments.append([a, b, c, d])

    invalid = Table('ID', 'LANGUAGE', 'CONCEPT', 'FORM')
    for row in analysis['stats']['invalid_words']:
        invalid.append(row)

    words = Table('ID', 'LANGUAGE', 'CONCEPT', 'FORM', 'SEGMENTS')
    for row in analysis['stats']['bad_words']:
        words.append(row)
    return TEMPLATE.format(
        segments.render(verbose=True),
        invalid.render(verbose=True),
        words.render(verbose=True))
